#%% LIBRARY IMPORT
import numpy as np
from scipy.optimize import minimize
from scipy.optimize import bisect

#%% CONSTANTS
E = 206*10**9 #Young's modulus of steel, in Pa
rho = 7850 #Iron/Steel density, in kg/m3
fk = 200*10**6 #Iron/Steel resistance, in N/m2

#%% DETERMINANT
# The following function evaluates the determinant of the matrix associated
# to the system that describes the free vibrations of an axially-loaded
# Euler-Bernoulli beam with 
def det_asym(T1, T2, a, b):
    piu = T1 + T2
    meno = T2 - T1
    L = 0.5*np.sqrt(-a + np.sqrt(a**2 + b**2))
    M = 0.5*np.sqrt(a + np.sqrt(a**2 + b**2))
    SL = np.sin(L)
    CL = np.cos(L)
    TM = np.tanh(M)
    Sasim = np.matrix([[-SL, 0., -TM, 0.],
                       [0., CL, 0., 1.],
                       [4*L**2*SL-piu*L*CL, meno*L*SL, 
                            -piu*M-4*M**2*TM, -meno*M*TM],
                       [meno*L*CL, -piu*L*SL-4*L**2*CL, 
                            meno*M, piu*M*TM+4*M**2]])
    return np.linalg.det(Sasim)

#%% FUNCTIONS FOR DET=0
def find_extremes(f, start, step):
    """
    Returns a list of two x-values between which a root of 
    *f(x)* is expected.   
    """
    x1 = start
    x2 = start + step
    
    while f(x1)*f(x2) > 0:
        x1 = x2
        x2 += step
    
    return [x1, x2]

def list_extremes(f, start, step, zeroes):
    """
    Invokes *find_extremes* repeatedly until it's found a
    list of *zeroes* roots for the function *f(x)*. Looks for
    roots at increasing distances.
    """
    extr0 = find_extremes(f, start, step)
    extr1 = find_extremes(f, extr0[1], step)
    ris = [extr0, extr1]
    if zeroes > 2:
        for i in range(2, zeroes):
            start_val = 2*ris[-1][1] - ris[-2][1]
            ris.append(find_extremes(f, start_val, step))
    return ris

def find_roots(f, zeroes=10, start=0.1, step=1.0):
    """
    Returns a list of *zeroes* roots for *f*.
    """
    look_here = list_extremes(f, start, step, zeroes)
    root_list = [bisect(f, extr[0], extr[1]) for extr in look_here]
    return root_list

#%% CONVERSION FUNCTIONS
def from_beta_to_freq(beta_list, tierod_dic):
    """
    Returns a list of frequencies corresponding to a list of beta values.
    *tierod_dic* = dictionary with tie-rod data (see README.md for contents)    
    """
    A = tierod_dic['area']
    J = tierod_dic['inertia']
    L = tierod_dic['length']
    molt = np.sqrt(rho*A/(E*J))*L**2*2*np.pi
    return beta_list/molt

def from_freq_to_beta(freq_list, tierod_dic):
    """
    Returns a list of beta values corresponding to a list of frequencies.
    *tierod_dic* = dictionary with tie-rod data (see README.md for contents)    
    """
    A = tierod_dic['area']
    J = tierod_dic['inertia']
    L = tierod_dic['length']
    molt = np.sqrt(rho*A/(E*J))*L**2*2*np.pi
    return freq_list*molt

def from_alpha_to_load(alpha, tierod_dic):
    """
    Returns the axial load corresponding to a given alpha value.
    *tierod_dic* = dictionary with tie-rod data (see README.md for contents)    
    """
    J = tierod_dic['inertia']
    L = tierod_dic['length']
    molt = 2.*(E*J)/L**2
    return alpha*molt

def from_load_to_alpha(axial_load, tierod_dic):
    """
    Returns an alpha value corresponding to a given axial load.
    *tierod_dic* = dictionary with tie-rod data (see README.md for contents)
    """
    J = tierod_dic['inertia']
    L = tierod_dic['length']
    molt = L**2/(2*E*J)
    return axial_load*molt

#%% DIRECT PROBLEM
def solve_direct_problem(tierod_dic, n_modes=10):
    """
    Reads axial load, boundary stiffness and other data from a tie-rod 
    dictionary and returns a list of *n_modes* frequencies.
    """
    N = tierod_dic['load']
    k1 = tierod_dic['k1']
    k2 = tierod_dic['k2']  
    alpha = from_load_to_alpha(N, tierod_dic)
    list_of_betas = find_roots(lambda b: det_asym(k1, k2, alpha, b), 
                               zeroes=n_modes)
    list_of_freqs = from_beta_to_freq(list_of_betas, tierod_dic)
    return list_of_freqs
   
#%% ERROR FUNCTION
def compute_error(measured_freq, theoretical_freq):
    """
    Returns the error calculated comparing experimental and 
    theoretical frequencies.
    """
    err = 0.0
    n_freq = 0
    for i in range(len(measured_freq)):
        mf = measured_freq[i]
        if mf != 'Null':
            tf = theoretical_freq[i]
            err += np.abs(mf - tf)/mf
            n_freq += 1
    return err/n_freq

#%% OBJECTIVE FUNCTIONS
def obj_func(x, tierod_dic, only_alpha=False, only_k=False):
    """
    Returns the error calculated for a tie-rod with:
    *x* = [alpha, k1, k2].
    If the switch *only_alpha* is on True, *x* = alpha and k-values
    are taken from *tierod_dic*.
    If the switch *only_k* is on True, *x* = [k1, k2] and the alpha 
    value is taken from *tierod_dic*.
    """
    if only_alpha:  #If constraint stiffness is given
        alpha = x
        k1 = tierod_dic['k1']
        k2 = tierod_dic['k2']
    
    elif only_k:  #If the load is given
        alpha = from_load_to_alpha(tierod_dic['load'], tierod_dic)
        k1 = x[0]
        k2 = x[1]
        
    else:
        alpha = x[0]
        k1 = x[1]
        k2 = x[2]
        
    meas_freq = tierod_dic['freq'] #This in any case
    
    if alpha < 0.0 or alpha >= 1000.: #Bounds for the optimization
        return 10**10
    
    th_beta = find_roots(lambda b: det_asym(k1, k2, alpha, b))
    th_freq = from_beta_to_freq(th_beta, tierod_dic)
    err = compute_error(meas_freq, th_freq)   
    return err

#%% FIRST ESTIMATE
def first_estimate(tierod_dic):
    """
    Returns a list [alpha0, k0] of first estimates starting from
    the ratio between measured frequencies and max/min theoretical
    frequencies.
    """
    n_modes = 1
    mf = tierod_dic['freq']
    A = tierod_dic['area']
    Nmax = A*fk
    
    min_tierod = tierod_dic.copy()
    min_tierod['load'] = 0.0
    min_tierod['k1'] = 0.0
    min_tierod['k2'] = 0.0
    max_tierod = tierod_dic.copy()
    max_tierod['load'] = Nmax
    max_tierod['k1'] = 10**10
    max_tierod['k2'] = 10**10
    
    tf_min = solve_direct_problem(min_tierod, n_modes=n_modes)
    tf_max = solve_direct_problem(max_tierod, n_modes=n_modes)
    ratio = (mf[0] - tf_min[0])/(tf_max[0] - tf_min[0])
    load0 = Nmax*ratio
    load_string = 'Load estimate = %.2f N' % load0
    
    print(load_string)
    
    k0 = 10.
    
    return [10., k0]

#%% OPTIMIZATION FUNCTION
def solve_inverse_problem(tierod_dic, fixed_load='Nope', fixed_k='Nope'):
    """
    Function that reads frequencies and other data from a tie-rod 
    dictionary and returns an estimate for the axial load.
    *tierod_dic* = dictionary with tie-rod data (see README.md for contents)
    If *fixed_load* is set to a value, it's used as axial load.
    Similarly, *fixed_k* can be set to a list of two values [k1, k2].
    """
    # Initial estimates for the parameters. Some may be
    # unused depending on the switches.
    zero = [10., 10., 10.]
    alpha_0 = zero[0]
    k1_0 = zero[1]
    k2_0 = zero[1]
    
    # Maxiter for minimizer function
    min_opt = {'maxiter': 10000}
    if fixed_load != 'Nope' and fixed_k != 'Nope':
        print('Maybe you wanted the direct problem?')
        return solve_direct_problem(tierod_dic)
    # Tollerance for minimizer function
    tollerance = 1e-10
    
    # If both the axial load and k are to be estimated
    if fixed_load == 'Nope' and fixed_k == 'Nope':
        x0 = [alpha_0, k1_0, k2_0]
        minimum = minimize(lambda x: obj_func(x, tierod_dic), 
                           x0, method='Nelder-Mead', tol=tollerance,
                           options=min_opt)
        load = from_alpha_to_load(minimum['x'][0], tierod_dic)
        k1 = minimum['x'][1]
        k2 = minimum['x'][2]
        
    # If only the constraints are requested
    elif fixed_load != 'Nope':
        tierod_dic['load'] = fixed_load
        x0 = [k1_0, k2_0] #Initial guess for stiffness not 0.0
        minimum = minimize(lambda x: obj_func(x, tierod_dic, only_k=True),
                           x0, method='Nelder-Mead', tol=tollerance,
                           options=min_opt)
        load = fixed_load
        k1 = minimum['x'][0]
        k2 = minimum['x'][1]
        
    # If only the load is requested
    elif fixed_k != 'Nope':
        tierod_dic['k1'] = fixed_k[0]
        tierod_dic['k2'] = fixed_k[1]
        x0 = alpha_0
        minimum = minimize(lambda x: obj_func(x, tierod_dic, 
                                              only_alpha=True), x0,
                            method='Nelder-Mead', tol=tollerance,
                            options=min_opt)
        load = minimum['x']
        k1 = fixed_k[0]
        k2 = fixed_k[1]
    
    err = minimum['fun']
    mess = minimum['message']   
    load_string = 'Load = %.2f N' % load
    if fixed_load != 'Nope':
        load_string += ' (FIXED)'
    k1_string = 'k1 = %.2f *EJ/L' % k1
    k2_string = 'k2 = %.2f *EJ/L' % k2
    if fixed_k != 'Nope':
        k1_string += ' (FIXED)'
        k2_string += ' (FIXED)'
    err_string = 'Error function = %.20f' % err    
    
    print('-'*30)
    print(str(tierod_dic['id']))
    print('-'*30)
    print(mess)
    print(load_string)
    print(k1_string)
    print(k2_string)
    print(err_string)
    print('\n')
    
    return [load, k1, k2, err]

#%% MISSING FREQUENCIES HEURISTIC
def missing_modes_heuristic(tierod_dic):
    """
    Calculates the maximum and minimum frequencies that could, in theory,
    correspond to a tie-rod of given characteristics. Then compares them with
    the measured ones to establish a range in which the number of missing
    modes falls. Finally, it creates a list of theoretical tie-rods to be
    tested, solves them with the indirect-problem solver, and prints the 
    results returning the solution to which the lowest error corresponds.
    """
    A = tierod_dic['area']
    Nmax = A*fk
    f_exp = tierod_dic['freq']
    print(f_exp)
    
    min_tierod = tierod_dic.copy()
    min_tierod['load'] = 0.0
    min_tierod['k1'] = 0.0
    min_tierod['k2'] = 0.0
    max_tierod = tierod_dic.copy()
    max_tierod['load'] = Nmax
    max_tierod['k1'] = 10**10
    max_tierod['k2'] = 10**10
    
    tf_min = solve_direct_problem(min_tierod, n_modes=10)
    print(tf_min)
    tf_max = solve_direct_problem(max_tierod, n_modes=10)
    print(tf_max)
    
    lim_sup = 0
    while tf_min[lim_sup] < f_exp[0]:
        lim_sup += 1
    lim_inf = 0
    while tf_max[lim_inf] < f_exp[0]:
        lim_inf += 1
    
    print(lim_inf)
    print(lim_sup)
    
    tierods_to_test = []    
    for i in range(lim_inf, lim_sup+1):
        foo = tierod_dic.copy()
        foo['freq'] = ['Null']*i + f_exp
        tierods_to_test.append(foo)
    print(tierods_to_test)

    test_loads = []
    test_errors = []
    for i, tr in enumerate(tierods_to_test):
        sol = solve_inverse_problem(tr)
        test_loads.append(sol[0])
        test_errors.append(sol[-1])
        print('FREQ: {}\n Load: {}kN \n Error: {}'.format(tr['freq'], 
              sol[0]/1000., sol[-1]))
    
    index_min = np.argmin(test_errors)
    
    return test_loads[index_min]
    
        
